import pyrealsense2 as rs
import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation as R


###########################################################
# # This program captures a multiple point clouds from the 
# # depth sensor from multiple different angles, 
# # filters out points that are too far away, 
# # and then saves it to disk as a .ply file. 
###########################################################

#####
# Constants
#####

#set max clipping distance
clipping_distance_in_meters = 0.4 #40cm clipping

# Distance between the far end of the axle and the back end of the sensor arm
# This should be changed as necessary with regards to FOV
axle_to_arm_distance = 0.378 #37.8cm
# distance between the front of the sensor and the back of the sensor arm
sensor_to_arm_distance = 0.048 #4.8cm
axle_outer_radius = 0.02 #2cm
# The real distance from the front of the sensor to the centerline of the axle
distance_to_center = axle_to_arm_distance - axle_outer_radius - sensor_to_arm_distance
center_vector = np.array((0,0,distance_to_center))



def get_depth_vectors (pipeline):
    #####
    # Takes in a realsense pipeline 
    # extracts a depth frame and then
    # gives out a numpy array of vectors for
    # each depth point in the depth frame.
    #####

    pc = 0


    #First we take a depth image with the sensor
    frames = pipeline.wait_for_frames()
    depth_frame = frames.get_depth_frame()

    #turn depth frame into RS pointcloud
    pcd = rs.pointcloud()
    points = pcd.calculate(depth_frame)
    
    #turn point cloud into numpy array for working with
    array = np.asanyarray(points.get_vertices()).copy()
    pc = array.view(np.float32).reshape((array.size, 3))

    return pc


def filter_depth (depth_vectors, depth):
    #####
    # Sets vectors in depth_vectors greater than depth to zero-vectors
    #####

    #the lengths of each vector
    distance = np.linalg.norm(depth_vectors, axis=1) 

    zeroVect = np.zeros((1,3)) #If a point has a value of 0 it's considered not a point and this can be used for filtering background
    distance_3d = np.dstack((distance, distance, distance))
    filtered_vectors = np.where(distance_3d <= depth, depth_vectors, zeroVect)
    filtered_vectors = filtered_vectors[0,:,:] #removes extra dimension that appears for some reason
    return filtered_vectors

def rotate_vector_about_z(vector, theta):
    #####
    # Rotates a vector in 3D space about the z-axis by theta radians
    #####
    r = R.from_rotvec(np.array((theta,0,0)))
    return r.apply(vector)

def get_centered_vectors(pipeline, clipping_distance, theta):
    #get teh vectors to be worked with
    vectors = get_depth_vectors(pipeline)
    
    #filter out vectors that have a magnitude greater than clipping distance
    filtered_vectors = filter_depth(vectors, clipping_distance)

    #Change the vectors' reference zero to center_vector as opposed to the sensor
    centered_vectors = np.subtract(filtered_vectors, center_vector)

    #rotate the vectors to fit into the desired reference frame
    rotated_vectors = rotate_vector_about_z(centered_vectors, theta)
    return rotated_vectors

#set up pipeline
pipe = rs.pipeline()
config = rs.config()

#depth sensor configuration
config.enable_stream(rs.stream.depth, 640, 480, rs.format.z16, 30)
#config.enable_stream(rs.stream.pose, rs.format.motion_xyz32f)

#start streaming
profile = pipe.start(config)

#colorizer for representation later
colorizer = rs.colorizer()

#set device preset, this allows for close range operation,
#but makes long range results less reliable
depth_sensor = profile.get_device().first_depth_sensor()
depth_sensor.set_option(rs.option.visual_preset, rs.l500_visual_preset.short_range)


numer = 0 #The starting multiple of pi
num_of_steps = 8 # each image is taken pi/4 rad apart for a total of 8 images

vecs = []
while numer <= num_of_steps/2:
    prompt = "Rotate to " + str(numer) + "pi/" + str(int(num_of_steps/2)) +" then hit enter"
    input(prompt)

    vectors = get_centered_vectors(pipe, clipping_distance_in_meters, numer * np.pi/4)
    vecs.append(vectors)
    numer+=1
vecs = np.array(vecs)

vecs = vecs.reshape((int(vecs.size/3), 3))

# Visualisation
pcd_out = o3d.geometry.PointCloud()
pcd_out.points = o3d.utility.Vector3dVector(vecs)
o3d.io.write_point_cloud("4.ply", pcd_out)
o3d.visualization.draw_geometries([pcd_out])

pipe.stop()
